@@ -1,11 +1,9 @@
-// $Id: SerialObj.h 6752 2009-06-14 04:24:52Z vern $
-//
 // Infrastructure for serializable objects.
 //
 // How to make objects of class Foo serializable:
 //
 //    1. Derive Foo (directly or indirectly) from SerialObj.
-//    2. Add a SER_FOO constant to SerialTypes below.
+//    2. Add a SER_FOO constant to SerialTypes in SerialTypes.h.
 //    3. Add DECLARE_SERIAL(Foo) into class definition.
 //    4. Add a (preferably protected) default ctor if it doesn't already exist.
 //    5. For non-abstract classes, add IMPLEMENT_SERIAL(Foo, SER_FOO) to *.cc
@@ -39,7 +37,7 @@
 #include "DebugLogger.h"
 #include "Continuation.h"
 #include "SerialTypes.h"
-#include "config.h"
+#include "bro-config.h"
 
 #if SIZEOF_LONG_LONG < 8
 # error "Serialization requires that sizeof(long long) is at least 8. (Remove this message only if you know what you're doing.)"
@@ -165,16 +163,16 @@ public:
 // Macro helpers.
 
 #define DECLARE_ABSTRACT_SERIAL(classname) \
-	virtual bool DoSerialize(SerialInfo*) const; \
-	virtual bool DoUnserialize(UnserialInfo*); \
+	bool DoSerialize(SerialInfo*) const override; \
+	bool DoUnserialize(UnserialInfo*) override; \
 
 #define DECLARE_SERIAL(classname) \
 	static classname* Instantiate(); \
 	static SerialTypeRegistrator register_type; \
-	virtual bool DoSerialize(SerialInfo*) const; \
-	virtual bool DoUnserialize(UnserialInfo*); \
-	virtual const TransientID*  GetTID() const	{ return &tid; } \
-	virtual SerialType GetSerialType() const; \
+	bool DoSerialize(SerialInfo*) const override; \
+	bool DoUnserialize(UnserialInfo*) override; \
+	const TransientID*  GetTID() const override	{ return &tid; } \
+	SerialType GetSerialType() const override; \
 	TransientID tid;
 
 // Only needed (and usable) for non-abstract classes.
@@ -321,7 +319,8 @@ public:
 	\
 	if ( has_it )	\
 		{	\
-		info->s->Read(&dst, 0, "has_" #dst);	\
+		if ( ! info->s->Read(&dst, 0, "has_" #dst) )	\
+			return false;	\
 		if ( ! dst )	\
 			return false;	\
 		}	\
@@ -330,6 +329,33 @@ public:
 		dst = 0;	\
 	}
 
+#define UNSERIALIZE_OPTIONAL_STR_DEL(dst, del)	\
+	{	\
+	bool has_it;	\
+	if ( ! info->s->Read(&has_it, "has_" #dst) )	\
+		{	\
+		delete del;	\
+		return 0;	\
+		}	\
+	\
+	if ( has_it )	\
+		{	\
+		if ( ! info->s->Read(&dst, 0, "has_" #dst) )	\
+			{	\
+			delete del;	\
+			return 0;	\
+			}	\
+		if ( ! dst )	\
+			{	\
+			delete del;	\
+			return 0;	\
+			}	\
+		}	\
+	\
+	else	\
+		dst = 0;	\
+	}
+
 #define UNSERIALIZE_OPTIONAL_STATIC(dst, unserialize, del)	\
 	{	\
 	bool has_it;	\
