@@ -1,13 +1,12 @@
-// $Id: Timer.cc 6219 2008-10-01 05:39:07Z vern $
-//
 // See the file "COPYING" in the main distribution directory for copyright.
 
-#include "config.h"
+#include "bro-config.h"
 
 #include "util.h"
 #include "Timer.h"
 #include "Desc.h"
 #include "Serializer.h"
+#include "broker/Manager.h"
 
 // Names of timers in same order than in TimerType.
 const char* TimerNames[] = {
@@ -18,11 +17,15 @@ const char* TimerNames[] = {
 	"ConnectionInactivityTimer",
 	"ConnectionStatusUpdateTimer",
 	"DNSExpireTimer",
+	"FileAnalysisInactivityTimer",
+	"FlowWeirdTimer",
 	"FragTimer",
 	"IncrementalSendTimer",
 	"IncrementalWriteTimer",
 	"InterconnTimer",
+	"IPTunnelInactivityTimer",
 	"NetbiosExpireTimer",
+	"NetWeirdTimer",
 	"NetworkTimer",
 	"NTPExpireTimer",
 	"ProfileTimer",
@@ -103,6 +106,7 @@ int TimerMgr::Advance(double arg_t, int max_expire)
 	last_timestamp = 0;
 	num_expired = 0;
 	last_advance = timer_mgr->Time();
+	broker_mgr->AdvanceTime(arg_t);
 
 	return DoAdvance(t, max_expire);
 	}
@@ -127,7 +131,7 @@ void PQ_TimerMgr::Add(Timer* timer)
 	// multiple already-added timers are added, they'll still
 	// execute in sorted order.
 	if ( ! q->Add(timer) )
-		internal_error("out of memory");
+		reporter->InternalError("out of memory");
 
 	++current_timers[timer->Type()];
 	}
@@ -173,7 +177,7 @@ int PQ_TimerMgr::DoAdvance(double new_t, int max_expire)
 void PQ_TimerMgr::Remove(Timer* timer)
 	{
 	if ( ! q->Remove(timer) )
-		internal_error("asked to remove a missing timer");
+		reporter->InternalError("asked to remove a missing timer");
 
 	--current_timers[timer->Type()];
 	delete timer;
@@ -183,7 +187,7 @@ CQ_TimerMgr::CQ_TimerMgr(const Tag& tag) : TimerMgr(tag)
 	{
 	cq = cq_init(60.0, 1.0);
 	if ( ! cq )
-		internal_error("could not initialize calendar queue");
+		reporter->InternalError("could not initialize calendar queue");
 	}
 
 CQ_TimerMgr::~CQ_TimerMgr()
@@ -208,7 +212,7 @@ void CQ_TimerMgr::Add(Timer* timer)
 		t = network_time;
 
 	if ( cq_enqueue(cq, t, timer) < 0 )
-		internal_error("problem queueing timer");
+		reporter->InternalError("problem queueing timer");
 
 	++current_timers[timer->Type()];
 	}
